;
;   Now some predefined things.  These buffers are smart enough to pass
;   delay information either way, but one or the other (or both) input
;   values must be known.
;
(declare (usual-integrations))
;
(def (buffer (a) (b))
   (if (wire-init-known? a)
      (when (not (wire-init-known? b))
         (wire-init-set! b (wire-init a)))
      (when (wire-init-known? b)
         (wire-init-set! a (wire-init b))))
   (when (not (= (wire-init a) (wire-init b)))
      (eval-parse (display-stuff " Pending event between <" (orb-eval 'a)
		     "> and <" (orb-eval 'b) ">" 'newline)))
   (if (= (wire-init a) 0)
      (if (= (wire-init b) 0)
         (process a+ (die-if a-) b+ a- (die-if a+) b-)
         (process (die-if a+) b- a+ (die-if a-) b+ a-))
      (if (= (wire-init b) 0)
         (process (die-if a-) b+ a- (die-if a+) b- a+)
         (process a- (die-if a+) b- a+ (die-if a-) b+))))
;
;   Now some predefined things.
;
(def (delay-buffer (a) (b) del)
   (if (wire-init-known? a)
      (when (not (wire-init-known? b))
         (wire-init-set! b (wire-init a)))
      (when (wire-init-known? b)
         (wire-init-set! a (wire-init b))))
   (when (not (= (wire-init a) (wire-init b)))
      (eval-parse (display-stuff " Pending event between <" (orb-eval 'a)
		     "> and <" (orb-eval 'b) ">" 'newline)))
   (if (= (wire-init a) 0)
      (if (= (wire-init b) 0)
         (process a+ (del) (die-if a- (ignore-timing)) b+ 
                  a- (del) (die-if a+ (ignore-timing)) b-)
         (process (del) (die-if a+ (ignore-timing)) b- a+
                  (del) (die-if a- (ignore-timing)) b+ a-))
      (if (= (wire-init b) 0)
         (process (del) (die-if a- (ignore-timing)) b+ a-
                  (del) (die-if a+ (ignore-timing)) b- a+)
         (process a- (del) (die-if a+ (ignore-timing)) b-
                  a+ (del) (die-if a- (ignore-timing)) b+))))
;
(def (delay-buffer-expand (a) (b) delup deldown)
   (if (wire-init-known? a)
      (when (not (wire-init-known? b))
         (wire-init-set! b (wire-init a)))
      (when (wire-init-known? b)
         (wire-init-set! a (wire-init b))))
   (when (not (= (wire-init a) (wire-init b)))
      (eval-parse (display-stuff " Pending event between <" (orb-eval 'a)
		     "> and <" (orb-eval 'b) ">" 'newline)))
   (if (= (wire-init a) 0)
      (if (= (wire-init b) 0)
         (process a+ (mydelay delup) (die-if a- (ignore-timing)) b+ 
                  a- (mydelay deldown) (die-if a+ (ignore-timing)) b-)
         (process (mydelay deldown) (die-if a+ (ignore-timing)) b- a+
                  (mydelay delup) (die-if a- (ignore-timing)) b+ a-))
      (if (= (wire-init b) 0)
         (process (mydelay delup) (die-if a- (ignore-timing)) b+ a-
                  (mydelay deldown) (die-if a+ (ignore-timing)) b- a+)
         (process a- (mydelay deldown) (die-if a+ (ignore-timing)) b-
                  a+ (mydelay delup) (die-if a- (ignore-timing)) b+))))
;
;   This `buffer' actually does very little but declare the wires
;   as receptive.
;
(def (receptive-buffer (a) (b))
   (if (wire-init-known? a)
      (wire-init-set! b (wire-init a))
      (when (wire-init-known? b)
         (wire-init-set! a (wire-init b))))
   (receptive-wire a b))
;
(def (default-delay-buffer (a) (b))
   (delay-buffer a b default-delay))
;
(def (si ckt)
   (connect buffer ckt #f))
;
(def (di ckt)
   (connect buffer ckt #t))
;
(def (receptive ckt)
   (connect receptive-buffer ckt #t))
;
(def (with-delay ckt)
   (connect delay-buffer ckt #f))
;
;   Now we define some standard combinational circuits.
;
(sm-ckt (fast-or (a b) (c)) (or a b))
(sm-ckt (fast-or3 (a b c) (d)) (or a (or b c)))
(sm-ckt (fast-nor (a b) (c)) (not (or a b)))
(sm-ckt (fast-nor3 (a b c) (d)) (not (or a (or b c))))
(sm-ckt (fast-and (a b) (c)) (and a b))
(sm-ckt (fast-and3 (a b c) (d)) (and a (and b c)))
(sm-ckt (fast-nand (a b) (c)) (not (and a b)))
(sm-ckt (fast-nand3 (a b c) (d)) (not (and a (and b c))))
(sm-ckt (fast-nand4 (a b c d) (e)) (not (and (and a d) (and b c))))
(sm-ckt (fast-not (a) (b)) (not a))
(sm-ckt (fast-buf (a) (b)) a)
(sm-ckt (fast-xor (a b) (c)) (not (= a b)))
(sm-ckt (fast-xnor (a b) (c)) (= a b))
(def (fast-c (a b) (c))
   (complete ()
      (wire-init-set! (a 0) (b 0) (c 0))
      (sm loop ((s 0))
         (guard
            ((= s 0) a+ (loop 1))
            ((= s 0) b+ (loop 2))
            ((= s 1) b+ :: c+ (loop 3))
            ((= s 1) a- (loop 0))
            ((= s 2) a+ :: c+ (loop 3))
            ((= s 2) b- (loop 0))
            ((= s 3) a- (loop 4))
            ((= s 3) b- (loop 5))
            ((= s 4) b- :: c- (loop 0))
            ((= s 4) a+ (loop 3))
            ((= s 5) a- :: c- (loop 0))
            ((= s 5) b+ (loop 3))))))
(def (fast-gc (a b) (c))
   (when (not (wire-init-known? c))
      (wire-init-set! c (wire-init a)))
   (complete ()
      (sm loop ((s (+ (wire-init a) (+ (* 2 (wire-init b))
                                        (* 4 (wire-init c))))))
         (case s
            (1 bad-gc-state)
            (3 a- (loop 2))
            (3 b- :: c+ (loop 5))
            (3 a- :: b- (loop 0))
            (6 bad-gc-state)
            (7 a- :: c- (loop 2))
            (7 b- (loop 5))
            (7 a- :: b- (loop 4))
	    (0 a+ :: c+ (loop 5))
	    (0 b+ (loop 2))
	    (0 a+ :: b+ (loop 3))
	    (2 a+ (loop 3))
	    (2 b- (loop 0))
	    (2 a+ :: b- :: c+ (loop 5))
	    (4 a+ (loop 5))
	    (4 b+ :: c- (loop 2))
	    (4 a+ :: b+ (loop 7))
	    (5 a- (loop 4))
	    (5 b+ (loop 7))
	    (5 a- :: b+ :: c- (loop 2))))))
;
(def (fast-latch (a b) (c))
   (if (wire-init a)
      (wire-init-set! c (wire-init a))
      (when (not (wire-init-known? c))
         (wire-init-set! c 0)))
   (complete ()
      (sm loop ((s (+ (* 4 (wire-init a))
                      (+ (* 2 (wire-init b)) (wire-init c)))))
         (guard
            ((= s 0) b+ (loop 2))
            ((= s 0) a+ (loop 4))
            ((= s 1) b+ (loop 3))
            ((= s 1) a+ :: c- (loop 4))
            ((= s 2) a+ :: c+ (loop 7))
            ((= s 2) b- (loop 0))
            ((= s 3) a+ (loop 7))
            ((= s 3) b- (loop 1))
            ((= s 4) b+ :: c+ (loop 7))
            ((= s 4) a- (loop 0))
            ((= s 7) a- (loop 3))
            ((= s 7) b- :: c- (loop 4))))))
;
;   We need to also have some two-phase gates.
;
(def (2buffer (a) (b))
   (if (wire-init-known? a)
      (when (not (wire-init-known? b))
         (wire-init-set! b (wire-init a)))
      (when (wire-init-known? b)
         (wire-init-set! a (wire-init b))))
   (process a+ (die-if a+) b+))
;
;   Now some predefined things.
;
(def (2delay-buffer (a) (b) del)
   (if (wire-init-known? a)
      (when (not (wire-init-known? b))
         (wire-init-set! b (wire-init a)))
      (when (wire-init-known? b)
         (wire-init-set! a (wire-init b))))
   (process a+ (del) (die-if a+ (ignore-timing)) b+))
;
(def (2fast-not (a) (b))
   (wire-init-set! b (not (wire-init a)))
   (process a+ :: b+))
;
(def (2fast-buf (a) (b))
   (wire-init-set! b (wire-init a))
   (process a+ :: b+))
;
(def (2fast-xor (a b) (c))
   (wire-init-set! c (if (= (wire-init a) (wire-init b)) 0 1))
   (if (= (wire-init c) 0)
      (process (choice (a+ :: c+) (b+ :: c+))
               (choice (a+ :: c-) (b+ :: c-)))
      (process (choice (a+ :: c-) (b+ :: c-))
               (choice (a+ :: c+) (b+ :: c+)))))
;
(def (2fast-xnor (a b) (c))
   (wire-init-set! c (if (= (wire-init a) (wire-init b)) 1 0))
   (if (= (wire-init c) 0)
      (process (choice (a+ :: c+) (b+ :: c+))
               (choice (a+ :: c-) (b+ :: c-)))
      (process (choice (a+ :: c-) (b+ :: c-))
               (choice (a+ :: c+) (b+ :: c+)))))
;
(def (2fast-latch (a b) (c))
   (wire-init-set! a 0)
   (when (wire-init-known? b)
      (when (not (wire-init-known? c))
         (wire-init-set! c (wire-init b))))
   (complete ()
      (sm loop ((s (if (= (wire-init a) (wire-init b)) 0 1)))
         (guard
            ((= s 0) b+ (loop 1))
            ((= s 1) b+ (loop 0))
            ((= s 0) a+ (loop 2))
            ((= s 1) a+ :: c+ (loop 2))
            ((= s 2) b+ :: c+ (loop 2))
            ((= s 2) a- (loop 0))))))
;
;   These are special values we use for timing
;
(def ignore -2)
;
(def unknown -1)
;
(def infinity -3)
;
(def inf -3)
